import type IVue from "vue";

interface RenderSlotItem {
  name: string;
  container: HTMLElement;
  slot: IVue.Slot;
  args: any;
}

export default function (Vue: typeof IVue) {
  const {
    h,
    ref,
    watch,
    toRef,
    reactive,
    computed,
    Teleport,
    onMounted,
    onBeforeUnmount,
    defineComponent,
  } = Vue;

  return function toVueComponent(HFC: typeof HyperFunctionComponent) {
    HFC.propTypes = HFC.propTypes || {};
    const propKeys = Object.keys(HFC.propTypes.attrs || {});
    const eventKeys = Object.keys(HFC.propTypes.events || {});
    const slotKeys = Object.keys(HFC.propTypes.slots || {});

    const eventKeyMap: Record<string, string> = {};
    for (let i = 0; i < eventKeys.length; i++) {
      const key = eventKeys[i];
      eventKeyMap["on" + key[0].toUpperCase() + key.slice(1)] = key;
    }

    const slotKeyMap: Record<string, string> = {};
    for (let i = 0; i < slotKeys.length; i++) {
      const key = slotKeys[i];
      slotKeyMap[key.toLowerCase()] = key;
    }

    const observeEventKeys = Object.keys(eventKeyMap);

    return defineComponent({
      props: propKeys,
      emits: eventKeys,
      setup(vueProps, { slots: vueSlots, attrs: vueAttrs }) {
        const container = ref<HTMLDivElement>();

        const events: Record<string, any> = {};
        if (observeEventKeys.length) {
          for (let i = 0; i < observeEventKeys.length; i++) {
            const key = observeEventKeys[i];
            const event = vueAttrs[key];
            if (!event) continue;
            events[eventKeyMap[key]] = event;
          }
        }

        const slots: Record<string, any> = {};
        const renderedSlots: Record<string, RenderSlotItem> = reactive({});

        Object.keys(vueSlots).forEach((slotKey) => {
          const key = slotKeyMap[slotKey.replace(/-/g, "").toLowerCase()];
          if (!key) return;

          const slot = vueSlots[slotKey];
          slots[key] = function (container: HTMLElement | null, args: any) {
            if (!container) {
              delete renderedSlots[key];
              return;
            }

            const slotItem = renderedSlots[key];
            if (!slotItem) {
              renderedSlots[key] = { name: key, container, slot: slot!, args };
              return;
            }

            if (slotItem.container !== container) {
              slotItem.container = container;
            }

            slotItem.args = args;
          };
        });

        let hfc: HyperFunctionComponent;
        const attrs: Record<string, any> = {};

        if (propKeys.length) {
          for (let i = 0; i < propKeys.length; i++) {
            const key = propKeys[i];
            if (vueProps[key] == undefined) continue;

            let value = vueProps[key];
            if (value === "") {
              const propType = HFC.propTypes!.attrs![key];
              if (propType.t === "#b") value = true;
            }

            attrs[key] = value;

            watch(
              toRef(vueProps, key),
              (curr, prev) => {
                attrs[key] = curr;
                hfc.changed?.("attr", key, prev, curr);
              },
              { deep: true }
            );
          }
        }

        hfc = new HFC({
          attrs,
          events,
          slots,
        });

        onMounted(() => {
          hfc.connected(container.value as HTMLDivElement);
        });

        onBeforeUnmount(() => {
          hfc.disconnected?.();
        });

        const renderedSlotVnodes = computed(() => {
          return Object.keys(renderedSlots).map((key) => {
            const item = renderedSlots[key];
            return h(Teleport, { to: item.container }, item.slot(item.args));
          });
        });

        return () => [h("div", { ref: container }), renderedSlotVnodes.value];
      },
    });
  };
}
